@@ -8,9 +8,9 @@ pybit is a lightweight and high-performance API connector for the
 RESTful and WebSocket APIs of the Bybit exchange.
 
 Documentation can be found at
-https://github.com/verata-veritatis/pybit
+https://github.com/bybit-exchange/pybit
 
-:copyright: (c) 2020-2021 verata-veritatis
+:copyright: (c) 2020-2022 bybit-exchange
 :license: MIT License
 
 """
@@ -19,9 +19,7 @@ import time
 import hmac
 import json
 import logging
-import threading
 import requests
-import websocket
 
 from datetime import datetime as dt
 from concurrent.futures import ThreadPoolExecutor
@@ -35,13 +33,17 @@ except ImportError:
     from json.decoder import JSONDecodeError
 
 # Versioning.
-VERSION = '1.3.4'
+VERSION = '2.2.0'
 
 
 class HTTP:
     """
     Connector for Bybit's HTTP API.
 
+    The below class is maintained for compatibility reasons. You should
+    prefer using the market-specific classes maintained in
+    inverse_perpetual.py, usdc_perpetual.py, spot.py, etc
+
     :param endpoint: The endpoint URL of the HTTP API, e.g.
         'https://api-testnet.bybit.com'.
     :type endpoint: str
@@ -121,6 +123,13 @@ class HTTP:
             handler.setLevel(logging_level)
             self.logger.addHandler(handler)
 
+        self.logger.warning(
+            'This HTTP class is maintained for compatibility purposes. You '
+            'should prefer importing market-specific classes, like so: '
+            'from pybit.inverse_perpetual import HTTP; '
+            'from pybit.spot import HTTP'
+        )
+
         self.logger.debug('Initializing HTTP session.')
         self.log_requests = log_requests
 
@@ -1066,12 +1075,15 @@ class HTTP:
         simultaneously.
 
         :param kwargs: See
+            https://bybit-exchange.github.io/docs/linear/#t-switchpositionmode,
             https://bybit-exchange.github.io/docs/inverse_futures/#t-switchpositionmode.
         :returns: Request results as dictionary.
         """
 
         if kwargs.get('symbol', '')[-2:].isdigit():
             suffix = '/futures/private/position/switch-mode'
+        elif kwargs.get('symbol', '').endswith('USDT'):
+            suffix = '/private/linear/position/switch-mode'
         else:
             suffix = '/v2/private/position/switch-mode'
 
@@ -1251,6 +1263,21 @@ class HTTP:
             auth=True
         )
 
+    def query_trading_fee_rate(self, **kwargs):
+        """
+        Query trading fee rate.
+
+        :param kwargs: See
+            https://bybit-exchange.github.io/docs/inverse/#t-queryfeerate.
+        :returns: Request results as dictionary.
+        """
+        return self._submit_request(
+            method='GET',
+            path=self.endpoint + '/v2/private/position/fee-rate',
+            query=kwargs,
+            auth=True
+        )
+
     def get_risk_limit(self, endpoint="", **kwargs):
         """
         Get risk limit.
@@ -1618,8 +1645,6 @@ class HTTP:
         """
         Create internal transfer.
 
-        :param kwargs: See
-            https://bybit-exchange.github.io/docs/account_asset/#t-querysubaccountlist.
         :returns: Request results as dictionary.
         """
 
@@ -1894,581 +1919,3 @@ class HTTP:
             else:
                 return s_json
 
-
-class WebSocket:
-    """
-    Connector for Bybit's WebSocket API.
-    """
-
-    def __init__(self, endpoint, api_key=None, api_secret=None,
-                 subscriptions=None, logging_level=logging.INFO,
-                 max_data_length=200, ping_interval=30, ping_timeout=10,
-                 restart_on_error=True, purge_on_fetch=True,
-                 trim_data=True):
-        """
-        Initializes the websocket session.
-
-        :param endpoint: Required parameter. The endpoint of the remote
-            websocket.
-        :param api_key: Your API key. Required for authenticated endpoints.
-            Defaults to None.
-        :param api_secret: Your API secret key. Required for authenticated
-            endpoints. Defaults to None.
-        :param subscriptions: A list of desired topics to subscribe to. See API
-            documentation for more information. Defaults to an empty list, which
-            will raise an error.
-        :param logging_level: The logging level of the built-in logger. Defaults
-            to logging.INFO. Options are CRITICAL (50), ERROR (40),
-            WARNING (30), INFO (20), DEBUG (10), or NOTSET (0).
-        :param max_data_length: The maximum number of rows for the stored
-            dataset. A smaller number will prevent performance or memory issues.
-        :param ping_interval: The number of seconds between each automated ping.
-        :param ping_timeout: The number of seconds to wait for 'pong' before an
-            Exception is raised.
-        :param restart_on_error: Whether or not the connection should restart on
-            error.
-        :param purge_on_fetch: Whether or not stored data should be purged each
-            fetch. For example, if the user subscribes to the 'trade' topic, and
-            fetches, should the data show all trade history up to the maximum
-            length or only get the data since the last fetch?
-        :param trim_data: Decide whether the returning data should be
-            trimmed to only provide the data value.
-
-        :returns: WebSocket session.
-        """
-
-        self.spot = True if "spot" in endpoint else False
-        self.spot_unauth = True if [True for v in ['v1', 'v2'] if v in
-                                    endpoint] else False
-        self.spot_auth = True if "spot" in endpoint and not \
-            self.spot_unauth else False
-        if not self.spot_auth:
-            if not subscriptions:
-                raise Exception('Subscription list cannot be empty!')
-        if not self.spot:
-            # Require symbol on 'trade' topic.
-            if 'trade' in subscriptions:
-                raise Exception('\'trade\' requires a ticker, e.g. '
-                                '\'trade.BTCUSD\'.')
-
-            # Require currency on 'insurance' topic.
-            if 'insurance' in subscriptions:
-                raise Exception('\'insurance\' requires a currency, e.g. '
-                                '\'insurance.BTC\'.')
-
-            # Require timeframe and ticker on 'klineV2' topic.
-            if 'klineV2' in subscriptions:
-                raise Exception('\'klineV2\' requires a timeframe and ticker, e.g.'
-                                ' \'klineV2.5.BTCUSD\'.')
-
-        # Check if subscriptions are in the correct format
-        if self.spot and not self.spot_auth:
-            for subscription in subscriptions.copy():
-                if isinstance(subscription, str):
-                    try:
-                        subscriptions.pop(subscriptions.index(subscription))
-                        subscriptions.append(json.loads(subscription))
-                    except JSONDecodeError:
-                        raise Exception('Spot subscriptions should be dicts, '
-                                        'or strings that are valid JSONs.')
-        elif not self.spot:
-            for subscription in subscriptions:
-                if not isinstance(subscription, str):
-                    raise Exception('Futures subscriptions should be strings.')
-
-            for subscription in subscriptions:
-                if ('v2' in endpoint and 'symbol' in subscription) or \
-                   ('v1' in endpoint and 'symbol' in subscription['params']):
-                    raise Exception('Cannot subscribe to v1 topics with v2 '
-                                    'endpoint, or vice versa.')
-
-        # set websocket name for logging purposes
-        self.wsName = 'Authenticated' if api_key else 'Non-Authenticated'
-
-        # Setup logger.
-        self.logger = logging.getLogger(__name__)
-
-        if len(logging.root.handlers) == 0:
-            # no handler on root logger set -> we add handler just for this logger to not mess with custom logic from outside
-            handler = logging.StreamHandler()
-            handler.setFormatter(logging.Formatter(fmt='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
-                                                   datefmt='%Y-%m-%d %H:%M:%S'
-                                                   )
-                                 )
-            handler.setLevel(logging_level)
-            self.logger.addHandler(handler)
-
-        self.logger.debug(f'Initializing {self.wsName} WebSocket.')
-
-        # Ensure authentication for private topics.
-        if not self.spot and any(i in subscriptions for i in [
-            'position',
-            'execution',
-            'order',
-            'stop_order',
-            'wallet'
-        ]) and api_key is None:
-            raise PermissionError('You must be authorized to use '
-                                  'private topics!')
-
-        # Set endpoint.
-        self.endpoint = endpoint
-
-        # Set API keys.
-        self.api_key = api_key
-        self.api_secret = api_secret
-
-        # Set topic subscriptions for WebSocket.
-        self.subscriptions = subscriptions
-        # Checking if using auth spot connection.
-        if '/spot/ws' in self.endpoint:
-            self.subscriptions = ['outboundAccountInfo', 'executionReport',
-                                  'ticketInfo']
-        self.max_length = max_data_length
-
-        # Set ping settings.
-        self.ping_interval = ping_interval
-        self.ping_timeout = ping_timeout
-
-        # Other optional data handling settings.
-        self.handle_error = restart_on_error
-        self.purge = purge_on_fetch
-        self.trim = trim_data
-
-        # Set initial state, initialize dictionary and connect.
-        self._reset()
-        self._connect(self.endpoint)
-
-    def fetch(self, topic):
-        """
-        Fetches data from the subscribed topic.
-
-        :param topic: Required parameter. The subscribed topic to poll.
-        :returns: Filtered data as dict.
-        """
-
-        if self.spot and self.spot_unauth:
-            topic = self.conform_topic(topic)
-        # If the topic given isn't in the initial subscribed list.
-        if topic not in self.subscriptions:
-            raise Exception(f"You aren\'t subscribed to the {topic} topic.")
-
-        # Pop all trade or execution data on each poll.
-        # don't pop order or stop_order data as we will lose valuable state
-        if any(i in topic for i in ['trade', 'execution']) \
-                and not topic.startswith('orderBook') and \
-                "executionReport" not in topic:
-            data = self.data[topic].copy()
-            if self.purge:
-                self.data[topic] = []
-            return data
-        else:
-            try:
-                return self.data[topic]
-            except KeyError:
-                return []
-
-    def ping(self):
-        """
-        Pings the remote server to test the connection. The status of the
-        connection can be monitored using ws.ping().
-        """
-
-        self.ws.send(json.dumps({'op': 'ping'}))
-
-    def exit(self):
-        """
-        Closes the websocket connection.
-        """
-
-        self.ws.close()
-        while self.ws.sock:
-            continue
-        self.exited = True
-
-    def _auth(self):
-        """
-        Authorize websocket connection.
-        """
-
-        # Generate expires.
-        expires = int((time.time() + 1) * 1000)
-
-        # Generate signature.
-        _val = f'GET/realtime{expires}'
-        signature = str(hmac.new(
-            bytes(self.api_secret, 'utf-8'),
-            bytes(_val, 'utf-8'), digestmod='sha256'
-        ).hexdigest())
-
-        # Authenticate with API.
-        self.ws.send(
-            json.dumps({
-                'op': 'auth',
-                'args': [self.api_key, expires, signature]
-            })
-        )
-
-    def _connect(self, url):
-        """
-        Open websocket in a thread.
-        """
-
-        self.ws = websocket.WebSocketApp(
-            url=url,
-            on_message=lambda ws, msg: self._on_message(msg),
-            on_close=self._on_close(),
-            on_open=self._on_open(),
-            on_error=lambda ws, err: self._on_error(err)
-        )
-
-        # Setup the thread running WebSocketApp.
-        self.wst = threading.Thread(target=lambda: self.ws.run_forever(
-            ping_interval=self.ping_interval,
-            ping_timeout=self.ping_timeout
-        ))
-
-        # Configure as daemon; start.
-        self.wst.daemon = True
-        self.wst.start()
-
-        # Attempt to connect for X seconds.
-        retries = 10
-        while retries > 0 and (not self.ws.sock or not self.ws.sock.connected):
-            retries -= 1
-            time.sleep(1)
-
-        # If connection was not successful, raise error.
-        if retries <= 0:
-            self.exit()
-            raise websocket.WebSocketTimeoutException('Connection failed.')
-
-        # If given an api_key, authenticate.
-        if self.api_key and self.api_secret and not self.spot_unauth:
-            self._auth()
-
-        # Check if subscriptions is a list.
-        if isinstance(self.subscriptions, (str, dict)):
-            self.subscriptions = [self.subscriptions]
-
-        # Subscribe to the requested topics.
-        if not self.spot_auth and self.spot_unauth:
-            for subscription in self.subscriptions:
-                if not subscription.get('event'):
-                    subscription['event'] = 'sub'
-                if not subscription.get('params'):
-                    subscription['params'] = {}
-                    if 'v2' in self.endpoint:
-                        raise Exception('v2 spot websocket topics require the '
-                                        '"symbol" key within "params"')
-                if not subscription.get('binary') or \
-                        subscription['params'].get('binary'):
-                    subscription['params']['binary'] = False
-                self.ws.send(json.dumps(subscription))
-        elif not self.spot:
-            self.ws.send(
-                json.dumps({
-                    'op': 'subscribe',
-                    'args': self.subscriptions
-                })
-            )
-
-        # Initialize the topics.
-        if not self.spot_auth and self.spot:
-            # Strip the subscription dict
-            for subscription in self.subscriptions:
-                index = self.subscriptions.index(subscription)
-                subscription = subscription if isinstance(subscription, dict) \
-                    else json.loads(subscription)
-                subscription.pop('event')
-                subscription['params']['binary'] = str(subscription['params'][
-                    'binary']).lower()
-                if subscription['params'].get('dumpScale'):
-                    subscription['params']['dumpScale'] = str(subscription[
-                        'params']['dumpScale'])
-                self.subscriptions[index] = \
-                    self.conform_topic(subscription)
-
-        topics = self.subscriptions
-        for topic in topics:
-            if topic not in self.data:
-                self.data[topic] = {}
-
-    @staticmethod
-    def _find_index(source, target, key):
-        """
-        Find the index in source list of the targeted ID.
-        """
-        return next(i for i, j in enumerate(source) if j[key] == target[key])
-
-    def _on_message(self, message):
-        """
-        Parse incoming messages. Similar structure to the
-        official WS connector.
-        """
-
-        # Load dict of message.
-        msg_json = json.loads(message)
-
-        # Did we receive a message regarding auth or subscription?
-        auth_message = True if isinstance(msg_json, dict) and \
-            (msg_json.get('auth') or
-             msg_json.get('request', {}).get('op') == 'auth') else False
-        subscription_message = True if isinstance(msg_json, dict) and \
-            ((msg_json.get('event') == 'sub' or msg_json.get('code')) or
-             msg_json.get('request', {}).get('op') == 'subscribe') else False
-
-        # Check auth
-        if auth_message:
-            # If we get successful futures/spot auth, notify user.
-            if msg_json.get('success') is True or \
-                    msg_json.get('auth') == 'success':
-                self.logger.debug('Authorization successful.')
-                self.auth = True
-            # If we get unsuccessful auth, notify user.
-            elif msg_json.get('auth') == 'fail' or \
-                    msg_json.get('success') is False:
-                self.logger.debug('Authorization failed. Please check your '
-                                  'API keys and restart.')
-
-        # Check subscription
-        if subscription_message:
-            # If we get successful futures/spot subscription, notify user.
-            if msg_json.get('success') is True or \
-                    msg_json.get('msg') == 'Success':
-                sub = msg_json['topic'] if self.spot else msg_json[
-                    'request']['args']
-                self.logger.debug(f'Subscription to {sub} successful.')
-            # Futures subscription fail
-            elif msg_json.get('success') is False:
-                response = msg_json['ret_msg']
-                if 'unknown topic' in response:
-                    self.logger.error('Couldn\'t subscribe to topic.'
-                                      f' Error: {response}.')
-            # Spot subscription fail
-            elif msg_json.get('code'):
-                self.logger.error('Couldn\'t subscribe to topic.'
-                                  f' Error code: {msg_json["code"]}.'
-                                  f' Error message: {msg_json.get("desc")}.')
-
-        elif 'topic' in msg_json:
-
-            if self.spot:
-                # Conform received topic data so that we can match with our
-                # subscribed topic
-                topic = self.conform_topic(msg_json.copy())
-            else:
-                topic = msg_json['topic']
-
-            # If incoming 'orderbookL2' data.
-            if 'orderBook' in topic:
-
-                # Make updates according to delta response.
-                if 'delta' in msg_json['type']:
-
-                    # Delete.
-                    for entry in msg_json['data']['delete']:
-                        index = self._find_index(self.data[topic], entry, 'id')
-                        self.data[topic].pop(index)
-
-                    # Update.
-                    for entry in msg_json['data']['update']:
-                        index = self._find_index(self.data[topic], entry, 'id')
-                        self.data[topic][index] = entry
-
-                    # Insert.
-                    for entry in msg_json['data']['insert']:
-                        self.data[topic].append(entry)
-
-                # Record the initial snapshot.
-                elif 'snapshot' in msg_json['type']:
-                    if 'order_book' in msg_json['data']:
-                        self.data[topic] = msg_json['data']['order_book'] if self.trim else msg_json
-                    else:
-                        self.data[topic] = msg_json['data'] if self.trim else msg_json
-                    #self.data[topic] = msg_json['data']
-
-            # If incoming 'diffDepth' data.
-            elif 'diffDepth' in topic:
-
-                book_sides = {'b': msg_json['data'][0]['b'],
-                              'a': msg_json['data'][0]['a']}
-
-                if not self.data[topic]:
-                    self.data[topic] = book_sides
-                    return
-
-                for side, entries in book_sides.items():
-                    for entry in entries:
-
-                        # Delete.
-                        if float(entry[1]) == 0:
-                            index = self._find_index(
-                                self.data[topic][side], entry, 0)
-                            self.data[topic][side].pop(index)
-                            continue
-
-                        # Insert.
-                        price_level_exists = entry[0] in \
-                            [level[0] for level in self.data[topic][side]]
-                        if not price_level_exists:
-                            self.data[topic][side].append(entry)
-                            continue
-
-                        # Update.
-                        qty_changed = entry[1] != next(
-                            level[1] for level in self.data[topic][side] if
-                            level[0] == entry[0])
-                        if price_level_exists and qty_changed:
-                            index = self._find_index(
-                                self.data[topic][side], entry, 0)
-                            self.data[topic][side][index] = entry
-                            continue
-
-            # For incoming 'order' and 'stop_order' data.
-            elif any(i in topic for i in ['order', 'stop_order']):
-
-                # record incoming data
-                for i in msg_json['data']:
-                    try:
-                        # update existing entries
-                        # temporary workaround for field anomaly in stop_order data
-                        ord_id = topic + '_id' if i['symbol'].endswith('USDT') else 'order_id'
-                        index = self._find_index(self.data[topic], i, ord_id)
-                        self.data[topic][index] = i
-                    except StopIteration:
-                        # Keep appending or create new list if not already created.
-                        try:
-                            self.data[topic].append(i)
-                        except AttributeError:
-                            self.data[topic] = msg_json['data']
-
-            # For incoming 'trade' and 'execution' data.
-            elif any(i in topic for i in ['trade', 'execution']):
-
-                # Keep appending or create new list if not already created.
-                try:
-                    trades = [msg_json['data']] if isinstance(
-                        msg_json['data'], dict) else msg_json['data']
-                    for i in trades:
-                        self.data[topic].append(i)
-                except AttributeError:
-                    self.data[topic] = msg_json['data']
-
-                # If list is too long, pop the first entry.
-                if len(self.data[topic]) > self.max_length:
-                    self.data[topic].pop(0)
-
-            # If incoming data is in a topic which only pushes messages in
-            # the snapshot format
-            elif any(i in topic for i in ['insurance', 'kline', 'wallet',
-                                          'candle', 'realtimes', '"depth"',
-                                          '"mergedDepth"', 'bookTicker']):
-
-                # Record incoming data.
-                if 'v2' in self.endpoint:
-                    self.data[topic] = msg_json['data'] if self.trim else msg_json
-                else:
-                    self.data[topic] = msg_json['data'][0] if self.trim else msg_json
-
-            # If incoming 'instrument_info' data.
-            elif 'instrument_info' in topic:
-
-                # Make updates according to delta response.
-                if 'delta' in msg_json['type']:
-                    for i in msg_json['data']['update'][0]:
-                        self.data[topic][i] = msg_json['data']['update'][0][i]
-
-                # Record the initial snapshot.
-                elif 'snapshot' in msg_json['type']:
-                    self.data[topic] = msg_json['data'] if self.trim else msg_json
-
-            # If incoming 'position' data.
-            elif 'position' in topic:
-
-                # Record incoming position data.
-                for p in msg_json['data']:
-
-                    # linear (USDT) positions have Buy|Sell side and
-                    # updates contain all USDT positions.
-                    # For linear tickers...
-                    if p['symbol'].endswith('USDT'):
-                        try:
-                            self.data[topic][p['symbol']][p['side']] = p
-                        # if side key hasn't been created yet...
-                        except KeyError:
-                            self.data[topic][p['symbol']] = {p['side']: p}
-
-                    # For non-linear tickers...
-                    else:
-                        self.data[topic][p['symbol']] = p
-
-        elif isinstance(msg_json, list):
-            for item in msg_json:
-                topic = item.get('e')
-                if topic == "outboundAccountInfo":
-                    self.data[topic] = item
-                elif any(i in topic for i in ['executionReport', 'ticketInfo']):
-                    # Keep appending or create new list if not already created.
-                    try:
-                        self.data[topic].append(item)
-                    except AttributeError:
-                        self.data[topic] = item
-                    self.data[topic] = item
-
-    def _on_error(self, error):
-        """
-        Exit on errors and raise exception, or attempt reconnect.
-        """
-
-        if not self.exited:
-            self.logger.error(f'WebSocket {self.wsName} encountered error: {error}.')
-            self.exit()
-
-        # Reconnect.
-        if self.handle_error:
-            self._reset()
-            self._connect(self.endpoint)
-
-    def _on_open(self):
-        """
-        Log WS open.
-        """
-        self.logger.debug(f'WebSocket {self.wsName} opened.')
-
-    def _on_close(self):
-        """
-        Log WS close.
-        """
-        self.logger.debug(f'WebSocket {self.wsName} closed.')
-
-    def _reset(self):
-        """
-        Set state booleans and initialize dictionary.
-        """
-        self.exited = False
-        self.auth = False
-        self.data = {}
-
-    @staticmethod
-    def conform_topic(topic):
-        """
-        For spot API. Due to the fact that the JSON received in update
-        messages does not include a simple "topic" key, and parameters all
-        have their own separate keys, we need to compare the entire JSON.
-        Therefore, we need to strip the JSON of any unnecessary keys,
-        cast some values, and dump the JSON with sort_keys.
-        """
-        if isinstance(topic, str):
-            topic = json.loads(topic)
-        topic.pop('symbolName', '')
-        topic['params'].pop('realtimeInterval', '')
-        topic['params'].pop('symbolName', '')
-        if topic['params'].get('klineType'):
-            topic['topic'] += "_" + topic['params'].get('klineType')
-            topic['params'].pop('klineType')
-        topic.pop('data', '')
-        topic.pop('f', '')
-        topic.pop('sendTime', '')
-        topic.pop('shared', '')
-        return json.dumps(topic, sort_keys=True, separators=(',', ':'))
